"""
imports added so users do not have to think about submodules
"""

from .core import describe_sizes, list_sizes, SaturnCluster  # noqa: F401
from ._version import get_versions
from .plugins import SaturnSetup, RegisterFiles, sync_files  # noqa: F401
from .external import ExternalConnection  # noqa: F401

__version__ = get_versions()["version"]
del get_versions

from . import _version
__version__ = _version.get_versions()['version']



"""
Worker plugins to be used with SaturnCluster objects.
"""
import os
import sys
import subprocess
from typing import List, Optional

from distributed import Client
from distributed.comm import resolve_address
from distributed.worker import get_client


PREFIX = "SATURN__"


class SaturnSetup:
    """WorkerPlugin that finishes setting up SaturnCluster."""

    name = "saturn_setup"

    def setup(self, worker=None):
        """This method gets called at worker setup for new and existing workers"""
        worker.scheduler.addr = resolve_address(worker.scheduler.addr)
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.append(cwd)


async def register_files_to_worker(paths: Optional[List[str]] = None) -> List[str]:
    """Register all files in the given paths on the current worker"""
    with get_client() as client:
        # If paths isn't provided, register all files in datasets that start with prefix
        if paths is None:
            datasets = await client.list_datasets()
            paths = [p[len(PREFIX) :] for p in datasets if p.startswith(PREFIX)]

        for path in paths:
            # retrieve the filedata from the scheduler
            payload = await client.get_dataset(f"{PREFIX}{path}")

            # by convention, paths that end with '/' are directories
            if path.endswith("/"):
                with open("/tmp/data.tar.gz", "wb+") as f:
                    f.write(payload)
                subprocess.run(f"mkdir -p {path}", shell=True, check=True)
                subprocess.run(f"tar -xvzf /tmp/data.tar.gz -C {path}", shell=True, check=True)
            else:
                basepath = os.path.split(path)[0]
                subprocess.run(f"mkdir -p {basepath}", shell=True, check=True)
                with open(path, "wb+") as f:
                    f.write(payload)
    return os.listdir()


def list_files(client: Client) -> List[str]:
    """List all files that are being tracked in the file registry"""
    datasets = client.list_datasets()
    return [p[len(PREFIX) :] for p in datasets if p.startswith(PREFIX)]


def clear_files(client: Client):
    """Clear all files that are being tracked in the file registry.

    After this is run, any new worker that is spun up, won't have any files
    automatically registered even if the RegisterFiles plugin is in use.
    """
    paths = list_files(client)
    for path in paths:
        client.unpublish_dataset(path)


def sync_files(client: Client, path: Optional[str] = None):
    """Upload files to all workers and add to file registry.

    :param client: distributed.Client object
    :param path: string or path obj pointing to file or directory to track.

    If used in conjunction with the ``RegisterFiles`` plugin, all files will be uploaded
    to new workers as they get spun up.
    """
    # normalize the path
    path = os.path.abspath(path)

    if os.path.isdir(path):
        path += "/"
        subprocess.run(
            f"tar --exclude .git -cvzf /tmp/data.tar.gz -C {path} .", shell=True, check=True
        )
        with open("/tmp/data.tar.gz", "rb") as f:
            payload = f.read()
    else:
        with open(path, "rb") as f:
            payload = f.read()

    # erase the given file or any file in the directory
    for p in [p for p in client.list_datasets() if path in p]:
        client.unpublish_dataset(p)

    client.publish_dataset(**{f"{PREFIX}{path}": payload})
    client.run(register_files_to_worker, paths=[path])


class RegisterFiles:
    """WorkerPlugin for uploading files or directories to dask workers.

    Use ``sync_files`` to control which paths are tracked.
    """

    name = "register_files"

    # pylint: disable=unused-argument
    async def setup(self, worker=None):
        """This method gets called at worker setup for new and existing workers"""
        await register_files_to_worker()



"""
Settings used for interacting with Saturn
"""

import os

from packaging.version import InvalidVersion, Version
from urllib.parse import urlparse
from ._version import get_versions

__version__ = get_versions()["version"]


class Settings:
    """Global settings"""

    SATURN_TOKEN: str
    SATURN_BASE_URL: str
    SATURN_VERSION: Version

    def __init__(self):
        try:
            self.SATURN_BASE_URL = os.environ["SATURN_BASE_URL"]
        except KeyError as err:
            if os.environ.get("BASE_URL") is not None:
                # if ``BASE_URL`` is set and ``SATURN_BASE_URL`` isn't, it's an old
                # version of Saturn that is incompatible with this version of dask_saturn.
                err_msg = (
                    "This version of dask-saturn is incompatible with your Saturn version. "
                    "Downgrade dask-saturn to `0.2.3`: `pip install dask_saturn==0.2.3`"
                )
                raise RuntimeError(err_msg) from err

            err_msg = "Missing required environment variable SATURN_BASE_URL."
            raise RuntimeError(err_msg) from err

        parsed = urlparse(self.SATURN_BASE_URL)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(f'"{self.SATURN_BASE_URL}" is not a valid URL')

        try:
            self.SATURN_TOKEN = os.environ["SATURN_TOKEN"]
        except KeyError as err:
            err_msg = "Missing required environment variable SATURN_TOKEN."
            raise RuntimeError(err_msg) from err

        # get the SATURN_VERSION if included, default to the one before field was added.
        version = os.environ.get("SATURN_VERSION", "2021.07.19")
        try:
            self.SATURN_VERSION = Version(version)
        except InvalidVersion:
            # suffix like `-daily` is not PEP440 compliant. Strip it off.
            self.SATURN_VERSION = Version(version.split("-")[0])

    @property
    def is_external(self) -> bool:
        """Whether the client environment is external to Saturn"""
        return os.environ.get("SATURN_IS_INTERNAL", "false").lower() == "false"

    @property
    def is_prefect(self) -> bool:
        """Whether the resource that the cluster will attach to is a prefect flow (or flow run)"""
        return os.environ.get("SATURN_RESOURCE_TYPE", "SingleUserServer").startswith("Prefect")

    @property
    def url(self):
        """Saturn url"""
        return self.SATURN_BASE_URL

    @property
    def headers(self):
        """Saturn auth headers"""
        return {
            "Authorization": f"token {self.SATURN_TOKEN}",
            "X-Dask-Saturn-Version": __version__,
        }



"""
Classes and functions for configuring the connection to a Saturn Dask cluster
from outside of the Saturn installation.
"""

import requests

from cryptography import x509
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from distributed.security import Security
from typing import List, Tuple
from urllib.parse import urljoin

from .settings import Settings


class ExternalConnection:
    """
    DEPRECATED: Stores settings for connecting to Saturn from an external location, and manages
    TLS certificates for communicating with the Dask Scheduler.
    """

    def __init__(self, *args, **kwargs):
        raise RuntimeError(
            "ExternalConnection is no longer supported. "
            "Instead, set the env vars: ``SATURN_TOKEN`` and ``SATURN_BASE_URL`` "
            "as indicated in the Saturn Cloud UI. If those env vars are set, an external "
            "connection will be automatically set up."
        )


def _client_tls(
    settings: Settings, dask_cluster_id: str
) -> (rsa.RSAPrivateKey, x509.Certificate, x509.Certificate):
    """
    Generate an RSA key and certificate signing request. Send the CSR to Saturn to be signed
    by the Dask cluster CA, and return the key, cert, and CA.
    """
    csr, key = _create_csr(f"Dask Client {dask_cluster_id}")

    url = urljoin(settings.url, f"/api/dask_clusters/{dask_cluster_id}/csr")
    resp = requests.post(
        url,
        data=_serialize_csr(csr),
        headers=settings.headers,
    )
    if not resp.ok:
        resp.raise_for_status()

    cert_chain = _deserialize_cert_chain(resp.content)
    if len(cert_chain) <= 1:
        raise ValueError("Invalid certificate chain returned from server")
    cert = cert_chain[0]
    ca_cert = cert_chain[1]

    return key, cert, ca_cert


def _security(settings: Settings, dask_cluster_id: str) -> Security:
    """
    Return Dask distributed security for connecting to the given dask cluster's scheduler
    over a public endpoint.
    """
    key, cert, ca_cert = _client_tls(settings, dask_cluster_id)
    key_contents = _serialize_key(key).decode()
    cert_contents = _serialize_cert(cert).decode()
    ca_cert_contents = _serialize_cert(ca_cert).decode()
    return Security(
        tls_client_key=key_contents,
        tls_client_cert=cert_contents,
        tls_ca_file=ca_cert_contents,
        require_encryption=True,
    )


def _create_csr(
    common_name: str,
) -> Tuple[x509.CertificateSigningRequest, rsa.RSAPrivateKeyWithSerialization]:
    """Return RSA key and certificate signing request"""
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COMMON_NAME, common_name),
                ]
            )
        )
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.ExtendedKeyUsage([ExtendedKeyUsageOID.CLIENT_AUTH]),
            critical=True,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(key.public_key()),
            critical=False,
        )
        .sign(key, hashes.SHA256(), default_backend())
    )
    return csr, key


def _serialize_csr(csr: x509.CertificateSigningRequest) -> bytes:
    """Return the PEM bytes from a certificate signing request"""
    return csr.public_bytes(serialization.Encoding.PEM)


def _serialize_cert(cert: x509.Certificate) -> bytes:
    """Return the PEM bytes from an X509 certificate"""
    return cert.public_bytes(serialization.Encoding.PEM)


def _serialize_key(key: rsa.RSAPrivateKeyWithSerialization) -> bytes:
    """Return the PEM bytes from an RSA private key"""
    return key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )


def _deserialize_cert_chain(data: bytes) -> List[x509.Certificate]:
    """Return the X509 certificate chain from the given PEM bytes"""
    pem_start_line = b"-----BEGIN CERTIFICATE-----\n"
    chain = data.split(pem_start_line)
    return [
        x509.load_pem_x509_certificate(pem_start_line + cert_bytes, default_backend())
        for cert_bytes in chain
        if cert_bytes
    ]



import os
from setuptools import setup, find_packages

import versioneer

install_requires = ["distributed", "packaging", "requests", "cryptography"]


setup(
    name="dask-saturn",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    maintainer="Saturn Cloud Developers",
    maintainer_email="dev@saturncloud.io",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="dask saturn cloud distributed cluster",
    description="Dask Cluster objects in Saturn Cloud",
    long_description=(open("README.md").read() if os.path.exists("README.md") else ""),
    long_description_content_type="text/markdown",
    url="https://saturncloud.io/",
    project_urls={
        "Documentation": "http://docs.saturncloud.io",
        "Source": "https://github.com/saturncloud/dask-saturn",
        "Issue Tracker": "https://github.com/saturncloud/dask-saturn/issues",
    },
    packages=find_packages(),
    package_data={"dask_saturn": ["*.yaml"]},
    install_requires=install_requires,
    tests_require=["pytest"],
    zip_safe=False,
)
